#!/usr/bin/ruby
require 'socket'
require 'thread'
require './common'

$port = ARGV.shift

class P4UServer
  def initialize
    @table = []
    @listen = TCPServer.open($port)
    puts("+ start server")
  end
  
  def main
    t = Thread.new {
      client_listen
    }
    local_session
    t.kill
  end
  
  def finalize
    @listen.close if @listen
    puts("+ stop server")
  end
  
  # NCAgZbV
  def client_listen
    while true do
      Thread.new(@listen.accept) {|s|
        client_session(s)
      }
    end
  end
  
  # NCAgZbV
  def client_session(sock)
    begin
      remote_host = sock.peeraddr[2]
      sock.puts(remote_host)
      puts("+ #{remote_host} connected")
      # 
      while line = sock.gets.chop do
        case line
        when /^ADD\s+(\S+)$/
          add_table(sock, $1)
          puts("+ ADD #{$1}")
        when /^DEL\s+(\S+)$/
          del_table(sock, $1)
          puts("+ DEL #{$1}")
        when /^LS\s+(\S.*)$/
          put_list(sock, $1)
          puts("+ LS #{$1}")
        when /^INFO\s+(\S+)$/
          put_info(sock, $1)
          puts("+ INFO #{$1}")
        else
          puts("! undefined operation")
        end
      end
    ensure
      puts("+ #{remote_host} leaved")
      sock.close
    end
  end
  
  # t@Cǉ
  def add_table(sock, data)
    info = data.to_info
    if @table.include?(info) then
      sock.puts("! #{info.name} already exist")
    else
      @table.push(info)
      sock.puts("+ #{info.name} added")
    end
  end
  
  # t@C폜
  def del_table(sock, data)
    info = data.to_info
    if @table.include?(info) then
      @table.delete(info)
      sock.puts("+ #{info.name} deleted")
    else
      sock.puts("! #{info.name} isn't exist")
    end
  end
  
  # ʂ𑗐M
  def put_list(sock, dir)
    # Mf[^
    array = @table.clone
    array.delete_if {|info|
      not info.name.include?(dir)
    }
    array.map! {|info|
      info.to_string
    }
    # f[^M
    sock.puts("#{array.size}")
    array.each {|info|
      sock.puts(info)
    }
  end
  
  # t@C𑗐M
  def put_info(sock, hash)
    # Mf[^
    array = @table.clone
    array.delete_if {|info|
      info.hash != hash
    }
    array.map! {|info|
      info.to_string
    }
    # f[^M
    sock.puts("#{array.size}")
    array.each {|info|
      sock.puts(info)
    }
  end
  
  # [JZbV
  def local_session
    while (line = gets) do
      case line.chop
      when /^quit$/
        break
      else
        puts("! undefined operation")
      end
    end
  end
end

begin
  server = P4UServer.new
  server.main
rescue
  puts("#{$!}")
ensure
  server.finalize
  exit(0)
end
